#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>

#include <netconf.h>
#include <nvparse.h>
#include "rc.h"
#include "usbmodem.h"


#define MODEM_DEVICES_FILE	"/proc/bus/usb/devices"
#define MODEM_DIAL_SCRIPT	"/usr/ppp/dial"
#define MODEM_DIAL_PIDFILE	"/var/run/dial%s.pid"

// see http://www.usb.org/developers/defined_class
#define USB_CLASS_PER_INTERFACE		0
#define USB_CLASS_COMM			2
#define USB_CLASS_MISC			0xef
#define USB_CLASS_VENDOR_SPEC		0xff

#define FORMAT_SHORT	1
#define FORMAT_JSON	2

#ifdef RC_SEMAPHORE_ENABLED
int hotplug_check_prev_zerocd_processed(const char *product, const char *device)
{
	char tmp[100], prefix[WAN_PREFIX_SZ];
	char str_devusb[100];
	int found = 0, unit;
	char *dev_vidpid;

	snprintf(str_devusb, sizeof(str_devusb), "zerocd: %s : %s", product, device);

	hotplug_sem_open();
	hotplug_sem_lock();

	// prevent multiple processing of the same device
	for (unit = 0; !found && unit < MAX_NVPARSE; unit ++)
	{
		snprintf(prefix, sizeof(prefix), "wan%d_", unit);

		dev_vidpid = nvram_get( strcat_r(prefix, "usb_device", tmp) );
		if (dev_vidpid)
			found = (strcmp(dev_vidpid, str_devusb) == 0);
#ifdef _DEBUG
		if (found) dprintf("already processed\n");
#endif
	}

	if (!found ) for (unit = 0; unit < MAX_NVPARSE; unit ++) {
		snprintf(prefix, sizeof(prefix), "wan%d_", unit);

		dev_vidpid = nvram_get( strcat_r(prefix, "usb_device", tmp) );
		if (!dev_vidpid) {
			dprintf("set: %s, %s\n", prefix, str_devusb);
			nvram_set( strcat_r(prefix, "usb_device", tmp), str_devusb );
			break;
		}
	}

	hotplug_sem_unlock();
	hotplug_sem_close();

	if (found) return 1;
	else return 0;
}

void hotplug_release_zerocd_processed(const char *product, const char *device)
{
	char tmp[100], prefix[WAN_PREFIX_SZ];
	char str_devusb[100];
	int unit;
	char *dev_vidpid;

	snprintf(str_devusb, sizeof(str_devusb), "zerocd: %s : %s", product, device);

	hotplug_sem_open();
	hotplug_sem_lock();

	// prevent multiple processing of the same device
	for (unit = 0; unit < MAX_NVPARSE; unit ++)
	{
		snprintf(prefix, sizeof(prefix), "wan%d_", unit);

		dev_vidpid = nvram_get( strcat_r(prefix, "usb_device", tmp) );
		if (dev_vidpid && strcmp(dev_vidpid, str_devusb) == 0) {
			nvram_unset( strcat_r(prefix, "usb_device", tmp) );
			break;
		}
	}

	hotplug_sem_unlock();
	hotplug_sem_close();
	dprintf("done: %s\n", device );
}
#else
#define hotplug_check_prev_zerocd_processed(product,device) 1
#define hotplug_release_zerocd_processed(product,device)
#endif /* RC_SEMAPHORE_ENABLED */

static void modem_get_script_name(char *filename)
{
	const char *script_fn = nvram_get("wan_modem_scriptfn");

	if (script_fn && *script_fn)
		strcpy(filename, script_fn);
	else
		strcpy(filename, MODEM_DIAL_SCRIPT);
};

/// get type and location for each device and count modems
/// with vid:pid (all if 0000:0000)
static int search_modems_in_list(dev_usb_t *list, int vid, int pid)
{
	dev_usb_t *dev = list;
	ifs_usb_elem_t *i;
	int count = 0;

	while (dev) {
		dprintf("dev:%04x:%04x\tprod:%s\tloc:%s\tcls:%x\tsub:%x\tprot:%x\n",
			dev->vid,dev->pid,dev->prod, dev->loc, dev->cls, dev->sub, dev->prot);
		// get type
		switch (dev->cls) {
			case USB_CLASS_COMM:
				dev->type = TYPE_WCDMA;
				if (vid || pid) {
					if (dev->vid == vid && dev->pid == pid)
						count++;
				} else {
					count++;
				}
				break;
			case USB_CLASS_MISC:
				// checking for Interface Association Descriptor ECN
				if (!(dev->sub == 0x02 && dev->prot == 0x01))
					break;
			case USB_CLASS_PER_INTERFACE:
				// class in IFACE
				for (i = dev->ifs; i; i = i->next) {
					if (i->cls == USB_CLASS_VENDOR_SPEC || i->cls == USB_CLASS_COMM) {
						dev->type = TYPE_WCDMA;
						if (vid || pid) {
							if (dev->vid == vid && dev->pid == pid)
							count++;
						} else {
							count++;
						}

						// empirical prediction of the port type!
						// The data port typically contains 3 EndPoints
						// one of which has interface with the
						// Interrupt transfers mode.
						if (i->endpoints > 2)
							dev->data_port = i->number;
						else
							dev->ui_port = i->number;
					}
				}
		}
		dev = dev->next;
	}
	return count;
}

/// check info in config file
static int check_config(dev_usb_t *dev)
{
	const dev_conf_t *conf;
	int i;
	for (i = 0; i < ARRAY_SIZE(modem_ids); i++) {
		conf = &modem_ids[i];
		// empty VID or type are not allowed
		if (!(conf->vid > 0 && conf->type > 0))
			continue;
		// VID (and PID if available) must match
		if (!(conf->vid == dev->vid && (conf->pid == -1 || conf->pid == dev->pid)))
			continue;
		// if we don't need to check interface info...
		if (conf->data_cls == -1 && conf->data_sub == -1 && conf->data_prot == -1 && conf->ui_cls == -1 && conf->ui_sub == -1 && conf->ui_prot == -1) {
			// ...check if least one port is specified...
			if (conf->data_port == -1 && conf->ui_port == -1)
				continue;
			// ...and accept the match
			dev->type = conf->type;
			if (conf->data_port != -1)
				dev->data_port = conf->data_port;
			if (conf->ui_port != -1)
				dev->ui_port = conf->ui_port;
			return 1;
		}
		// at least one class, subclass or protocol specified -- check
		else {
			ifs_usb_elem_t *i;
			int count = 0;
			// iterate over interfaces
			for (i = dev->ifs; i; i = i->next) {
				// check data interface
				if (conf->data_cls != -1 || conf->data_sub != -1 || conf->data_prot != -1) {
					if ((conf->data_cls == -1 || conf->data_cls == i->cls) && (conf->data_sub == -1 || conf->data_sub == i->sub) && (conf->data_prot == -1 || conf->data_prot == i->prot)
					&& (conf->data_port == -1 || conf->data_port == i->number)) {
						dev->type = conf->type;
						dev->data_port = i->number;
						count++;
					}
				}
				// check ui interface
				if (conf->ui_cls != -1 || conf->ui_sub != -1 || conf->ui_prot != -1) {
					if ((conf->ui_cls == -1 || conf->ui_cls == i->cls) && (conf->ui_sub == -1 || conf->ui_sub == i->sub) && (conf->ui_prot == -1 || conf->ui_prot == i->prot)
					&& (conf->ui_port == -1 || conf->ui_port == i->number)) {
						dev->type = conf->type;
						dev->ui_port = i->number;
						count++;
					}
				}
			}
			// at least one interface found
			if (count)
				return 1;
		}
	}
	return 0;
}

/// print info about modems
static void print_modem_info(int format, dev_usb_t *dev_list)
{
	dev_usb_t *dev;

	if (format == FORMAT_JSON) puts("[");

	for (dev = dev_list; dev; dev = dev->next) {
		if (dev->type == TYPE_CDMA || dev->type == TYPE_WCDMA)
			switch (format) {
			case FORMAT_SHORT:
				printf("%s:%s:%s\n",
					dev->loc, dev->manuf, dev->prod);
				break;

			case FORMAT_JSON:
				printf(	"{'wan':'0','vid':'%04x','pid':'%04x','type':'%c',"
					"'data':'%d','ui':'%d','loc':'%s','manuf':'%s','prod':'%s'},\n",
					dev->vid, dev->pid, dev->type, dev->data_port, dev->ui_port,
					dev->loc, dev->manuf, dev->prod);
				break;

			default:
				printf("0:%04x:%04x:%c:%d:%d:%s:%s:%s\n",
					dev->vid, dev->pid, dev->type, dev->data_port, dev->ui_port,
					dev->loc, dev->manuf, dev->prod);
		}
	}
	if (format == FORMAT_JSON) puts("]");
}

/// get usb-device number and usb-bus from environment DEVICE
int get_bus_num_from_devstr(const char *device, int *bus, int *num)
{
	char *ptr;
	int result = 0;

	ptr = strrchr (device, '/');
	if (ptr > device) {
		*num = atoi(ptr+1);
		ptr--;
		while (ptr >= device && *ptr != '/')
			ptr--;

		if (ptr > device) {
			*bus = atoi(ptr+1);

			dprintf("device: %s\t%d\t%d\n", device, *bus, *num);

			result = 1;
		}
	}
	
	return result;
}

/// get usb-device by environment DEVICE
static dev_usb_t *get_usb_by_location_env(const char *device, dev_usb_t *list)
{
	int bus, dev_num;
	dev_usb_t *usb_dev, *result;

	result = NULL;

	dprintf("%s\n",device);
	if (get_bus_num_from_devstr(device, &bus, &dev_num)) {
		for (usb_dev = list; usb_dev; usb_dev=usb_dev->next) {
			dprintf("%x/%x	%s	%d/%d\n", 
				usb_dev->vid, usb_dev->pid, usb_dev->loc,
				usb_dev->bus, usb_dev->number);

			if (usb_dev->bus == bus && usb_dev->number == dev_num &&
				(usb_dev->type == 'W' || usb_dev->type == 'C')) {
				result = usb_dev;
				break;
			}
		}
	}
	dprintf("done\n");
	return result;
}

int start_modem_dial(const char *prefix)
{
	int ret = 0;
	pid_t pid;
	FILE * file;
	char sfn[200], tmp[200];
	char *dial_argv[] = {
		(modem_get_script_name(sfn), sfn),
		(char *)prefix,
		NULL
	};

	if (nvram_match(strcat_r(prefix, "dial_enabled", tmp), "1") &&
		nvram_match(strcat_r(prefix, "prepared", tmp), "1") &&
		nvram_get(strcat_r(prefix, "usb_device", tmp))) {
		dprintf("%s\n", sfn);

		if (nvram_match(strcat_r(prefix, "proto", tmp), "usbmodem")) {
			nvram_set(strcat_r(prefix, "ifname", tmp) , nvram_safe_get(strcat_r(prefix, "pppoe_ifname", tmp)));
			nvram_set(strcat_r(prefix, "dnsenable_x", tmp), "1");
		}
		nvram_set("wan_status_t", "Connecting...");
		ret = _eval(dial_argv, NULL, 0, &pid);
		if (pid) {
			sprintf(tmp, MODEM_DIAL_PIDFILE, prefix);
			file = fopen(tmp, "w");
			if (file) {
				fprintf(file, "%d", pid);
				fclose(file);
			}
		}
	}

	dprintf("done\n");
	return ret;
}

int stop_modem_dial(const char *prefix)
{
	int ret;
	char tmp[100];
	int unit;


	unit = nvram_get_int(strcat_r(prefix, "unit", tmp));
	ret = 0;

	sprintf(tmp, MODEM_DIAL_PIDFILE, prefix);
	dprintf("kill %s\n", tmp);
	kill_pidfile(tmp);
	unlink(tmp);

	sprintf(tmp, "/var/run/ppp%d.pid", unit);
	dprintf("kill %s\n", tmp);
	kill_pidfile(tmp);
	//		nvram_set("wan0_dial_enabled", "0");

	dprintf("%s done\n", prefix);
	return ret;
}

int parse_product_string(const char *product, int *vid, int *pid)
{
	int res = 0;
	char tmp[512], *str1, *str2;
	*vid = 0; *pid = 0;

	strncpy(tmp, product, sizeof(tmp)-1);
	tmp[sizeof(tmp)-1] = '\0';
	str1 = strchr(tmp, '/');
	if (str1) {
		*str1 = 0; str1++;
		str2 = strchr(str1, '/');
		if (str2)
			*str2 = '\0';
		sscanf(tmp, "%x", vid);
		sscanf(str1, "%x", pid);
		res = 1;
	}
	return res;
}

static void modem_load_drivers()
{
	insmod_cond("cdc-acm", NULL);
	insmod_cond("usbserial", NULL);
	insmod_cond("option", NULL);
}

/// wait for device appearance in /proc/bus/usb/devices
int wait_for_dev_appearance(int vid, int pid, const char *device)
{
	dev_usb_t *dev, *list;
	int result, rep_counter, dev_bus, dev_num;

	const int max_time = 10;  //max wait in seconds
	const int rep_time = 100; //iteration time in milliseconds

	result = 0;
	rep_counter = max_time*1000/rep_time;

	if (!get_bus_num_from_devstr(device, &dev_bus, &dev_num))
		return result;

	while (!result && rep_counter) {
		list = usb_get_list();
		for (dev = list; dev; dev = dev->next) {
			if (dev->vid == vid && dev->pid == pid &&
			    dev->number == dev_num && dev->bus == dev_bus)
				result = 1;
		}
		usb_free_list(list);
		if (!result)
			usleep(rep_time*1000);
		rep_counter--;
	}
#ifdef DEBUG
	dprintf("rep_counter %d\n", max_time*1000/rep_time-rep_counter);
#endif

	return result;
}

int hotplug_check_modem(const char *interface, const char *product, const char *device, const char *prefix)
{
	int ret = 0, i;
	int count = 0;
	char *str1, *str2;
	char tmp[200];
	char stored_product[40];
	int vid, pid, hp_vid, hp_pid;
	dev_usb_t *list, *found_dev;

	int autodetect = nvram_match(strcat_r(prefix, "modem_autodetect", tmp), "1");
	//int autodetect = nvram_match("wan_modem_autodetect", "1");

	// if device was already found for this wan, do nothing
	if (nvram_get(strcat_r(prefix, "usb_device", tmp))) return 0;

	str1 = nvram_prefix_get(prefix, "modem_vid");
	//str1 = nvram_safe_get("wan_modem_vid");
	sscanf(str1, "%x", &vid);
	str2 = nvram_prefix_get(prefix, "modem_pid");
	//str2 = nvram_safe_get("wan_modem_pid");
	sscanf(str2, "%x", &pid);

	dprintf("stored: %x/%x, found: %s, autodetect: %d\n", vid, pid, product, autodetect);

	sscanf(product, "%x/%x", &hp_vid, &hp_pid);

	if (!wait_for_dev_appearance(hp_vid, hp_pid, device)) {
		dprintf("device not found in /proc/bus/usb/devices\n");
		return 0;
	}

	found_dev = NULL;

	list = usb_get_list();
	if (autodetect == 0) {
		if (vid == hp_vid && pid == hp_pid &&
			search_modems_in_list(list, 0, 0) > 0) {
				str1 = nvram_prefix_get(prefix, "modem_usbloc");
				//str1 = nvram_safe_get("wan_modem_usbloc");
				found_dev = get_usb_by_location_env(device, list);
#ifdef DEBUG
				if (found_dev)
				    dprintf("found loc=%s, vid=%04x, pid=%04x\n", found_dev->loc, found_dev->vid, found_dev->pid);
#endif
				if (found_dev && strcmp(str1, found_dev->loc) == 0 &&
					found_dev->vid == vid && found_dev->pid == pid)
					ret = 1;
		}
	} else {
		if (search_modems_in_list(list, 0, 0) > 0) {
			for (found_dev = list; found_dev; found_dev = found_dev->next) {
				check_config(found_dev);
			}

			if ((found_dev = get_usb_by_location_env(device, list))) {
				sprintf(stored_product, "0x%04x", found_dev->vid);
				nvram_set(strcat_r(prefix, "modem_vid", tmp), stored_product);
				//nvram_set("wan_modem_vid", stored_product);
				sprintf(stored_product, "0x%04x", found_dev->pid);
				nvram_set(strcat_r(prefix, "modem_pid", tmp), stored_product);
				//nvram_set("wan_modem_pid", stored_product);

				strcpy(stored_product, found_dev->loc);
				nvram_set(strcat_r(prefix, "modem_usbloc", tmp), stored_product);
				//nvram_set("wan_modem_usbloc", stored_product);

				sprintf(stored_product, "%d", found_dev->data_port);
				nvram_set(strcat_r(prefix, "modem_pdata", tmp), stored_product);
				//nvram_set("wan_modem_pdata", stored_product);

				sprintf(stored_product, "%d", found_dev->ui_port);
				nvram_set(strcat_r(prefix, "modem_pui", tmp), stored_product);
				//nvram_set("wan_modem_pui", stored_product);
				
				sprintf(stored_product, "%c", found_dev->type);
				nvram_set(strcat_r(prefix, "modem_type", tmp), stored_product);
				//nvram_set("wan_modem_type", stored_product);

				dprintf("wrote: %04x:%04x USB:%s\n", found_dev->vid, found_dev->pid, found_dev->loc);
				ret = 1;
			}
		}
	}

	if (ret && found_dev) {
		nvram_set(strcat_r(prefix, "usb_device_name", tmp), found_dev->prod);
		modem_load_drivers();
		
		for (i = 0; i < 10 && !count; i++) {
			usleep(500000);

			dev_usb_t *dev_list = usb_get_list();

			dev_usb_t *dev;
			ifs_usb_elem_t *i;

			for (dev = dev_list; dev; dev = dev->next) {
				if (!strcmp(dev->loc, found_dev->loc)) {
				for (i = dev->ifs; i; i = i->next) {
					if (!strcmp(i->driver, "cdc_acm") ||
					    !strcmp(i->driver, "option"))
						count++;
				}
			}
			}
			
			usb_free_list(dev_list);
		}
		
		if (!count) {
			FILE * fnew_id = fopen("/sys/bus/usb-serial/drivers/option1/new_id", "a");

			if (fnew_id) {
				fprintf(fnew_id, "0x%x 0x%x\n", found_dev->vid, found_dev->pid);
				fclose(fnew_id);
			} else {
				dprintf( "Option file open error: %s\n", strerror(errno) );
			}
		}
	}

	usb_free_list(list);

	dprintf("done. ret %d\n", ret);

	return ret;
}

void
hotplug_exec_user_modem_init_script(const char *sVid, const char *sPid, const char *sDevice)
{
	char **argv_user;
	char *tok, *context;
	const char *delim = " ";
	int i;
	char tmp[256];

	tok = nvram_get("wan_modem_init_script");
	if (tok && *tok) {
		strncpy(tmp, tok, sizeof(tmp) - 1);
		dprintf("prepare to execute %s\n", tmp);

		argv_user = (char **)malloc((strlen(tmp)/2) * sizeof(char*));
		tok = strtok_r(tmp, delim, &context);
		for (i = 0; tok ; i++, (tok = strtok_r(NULL, delim, &context))) {
			if (!strcasecmp(tok, "$VID"))
				argv_user[i] = (char *)sVid;
			else if (!strcasecmp(tok, "$PID"))
				argv_user[i] = (char *)sPid;
			else if (!strcasecmp(tok, "$DEV"))
				argv_user[i] = (char *)sDevice;
			else
				argv_user[i] = tok;
		};
		argv_user[i] = NULL;
		_eval(argv_user, NULL, 0, NULL);
		free((void*)argv_user);
		dprintf("%s executed\n", tmp);
	}
}

void hotplug_usb_modeswitch(const char *interface, const char *action, const char *product, const char *device)
{
	int vid, pid;

	int dev_vid, dev_pid;
	int i = 0;
	char sFileName[128];
	char sPath[128] = "";
	char sVid[0x0F] = "";
	char sPid[0x0F] = "";

#define CNT_VE_STR 2
	char *sMaList[] = {
		// names for sVe=...
		"Option",
		"GT",

		// names for uMa=...
		"AnyDATA",
		"DGT",
		"SAMSUNG",
		"Vertex",
		"SSE",
		"Philips",
		"Wisue"
	};

	char *argv_usb_modeswitch[] = {
		"usb_modeswitch",
#ifdef DEBUG
		"-W",
#endif
		"-D",
		"-v",
		sVid,
		"-p",
		sPid,
		"-c",
		sPath,
		NULL
	};

	if (strcmp(action, "add") == 0 && !hotplug_check_prev_zerocd_processed(product,device)) {

		if (parse_product_string(product, &vid, &pid)) {
			sprintf(sVid, "0x%04x", vid);
			sprintf(sPid, "0x%04x", pid);
		}

		if (nvram_match("wan_modem_zerocd_mode", "UserDefined")) {
			strcpy(sPath, "/usr/local/etc/usb_modeswitch.conf");
		} else if (nvram_match("wan_modem_zerocd_mode", "Auto")) {
			if ( *sVid && *sPid ){
				sprintf(sFileName, "%04x:%04x", vid, pid);
				if ((vid == 0x05c6 && pid == 0x1000) ||
					(vid == 0x1a8d && pid == 0x1000) ||
					(vid == 0x0471 && pid == 0x1210))
				{
					FILE *file;
					char str[255];
					int i_size;
					char * str1;
					int added = 0;

					file = fopen("/proc/bus/usb/devices", "rt");
					if (file) {
						int ready = 0;
						char *templ[] = {
							"Vendor=",
							"ProdID=",
							"Manufacturer=",
							"Product="};
						char sManufacturer[0xFF];

						while (!feof(file)) {
							fgets(str, sizeof(str)-1, file);
							if ((str1 = strstr(str, templ[0]))) {
								sscanf(str1+strlen(templ[0]), "%x", &dev_vid);
								if ((str1 = strstr(str1, templ[1]))) {
									sscanf(str1+strlen(templ[1]), "%x", &dev_pid);
								}
								ready = 0;
							} else if ((str1 = strstr(str, templ[2]))) {
								strncpy(sManufacturer, str1+strlen(templ[2]), sizeof(sManufacturer)-1);
								ready = 1;
							} else if ((str1 = strstr(str, templ[3]))) {
								strncpy(sManufacturer, str1+strlen(templ[3]), sizeof(sManufacturer)-1);
								ready = 2;
							}

							if (ready) {
								dprintf("lsUSB: %04x:%04x\n", dev_vid, dev_pid);
								if (dev_vid == vid && dev_pid == pid) {
									i_size = ARRAY_SIZE(sMaList);
									if (ready == 1) {
										for (i = 0; i < i_size; i++) {
											if (strncmp(sManufacturer, sMaList[i], strlen(sMaList[i])) == 0) {
												if (i < CNT_VE_STR) strcat (sFileName, ":sVe=");
												else strcat (sFileName, ":uMa=");

												strcat(sFileName, sMaList[i]);
												added = 1;
												break;
											}
										}
									} else {
										if (strncmp(sManufacturer, "5G", 2) == 0) {
											strcat (sFileName, ":uPr=5G");
											added = 1;
											break;
										}
									}
								}
								ready = 0;
							}
						}
						fclose(file);
					}
					if (added == 0 && vid == 0x05c6)
						strcat (sFileName, ":uMa=AnyDATA");
				}
			};
			sprintf(sPath, "/usr/share/usb_modeswitch.d/%s", sFileName);
		}

		if (*sPath) {
			dprintf("Check usb_modeswitch data file %s\n",sPath);
			if (exists(sPath)) {
				dprintf("usb_modeswitch -c %s\n", sPath);
				_eval(argv_usb_modeswitch, ">>/tmp/chat.log", 0, NULL);
			}
		}

		//******* exec user script ******************
		hotplug_exec_user_modem_init_script(sVid, sPid, device);

	} else if(strcmp(action, "remove") == 0 ) {
		hotplug_release_zerocd_processed(product, device);
	}
	dprintf("done\n");
}

int usb_modem_check(const char *prefix)
{
	int ret = 0;
	char tmp[100];

	if (nvram_match(strcat_r(prefix, "dial_enabled", tmp), "1") &&
		nvram_match(strcat_r(prefix, "prepared", tmp), "1") &&
		nvram_get(strcat_r(prefix, "usb_device", tmp)))
	{
		sprintf(tmp, MODEM_DIAL_PIDFILE, prefix);
		if (!proc_check_pid(tmp))
			ret = start_modem_dial(prefix);
	}
	return ret;
}

/// list modems ( see format in usbmodem.h )
/// usage: lsmodem [-s|-j]
///   -s  short form listing ( : location : manufacturer : product : )
///
/// modem description string
///  0:xxxx:xxxx:C:0:0:0.1.2:manufacturer:product
///  | |    |    | | | |     |            |
///  | |    |    | | | |     |             `usb product string (char[])
///  | |    |    | | | |      `usb manufacturer string (char[])
///  | |    |    | | |  `usb device location string (char[])
///  | |    |    | |  `user interface port for modem controlling, status sms, ussd and so on (char[1])
///  | |    |    |  `main port for data communication number (char[1])
///  | |    |     `modem type - 'C' for CDMA or 'W' for WCDMA/HSPA/HSDPA/HSUPA/GSM (char[1])
///  | |     `product id in hex (char[4])
///  |  `vendor id in hex (char[4])
///   `connection enable flag ( 0 - disable, 1 - enable ) (char[1])
int lsmodem_main(int argc, char **argv)
{
	int format = 0;
	int i;
	dev_usb_t *dev, *list;

	if (argc > 1) {
		for (i = 1; i < argc; i++) {
			if (strcmp(argv[i], "-s") == 0){ 
				format = FORMAT_SHORT;
			} else if (strcmp(argv[i], "-j") == 0) {
				format = FORMAT_JSON;
			} else {
				fprintf(stderr,"Usage: lsmodem [-s|-j]\n");
				return 1;
			}
		}
	}

	list = usb_get_list();
	if (search_modems_in_list(list, 0, 0) > 0 && format != FORMAT_SHORT) {
		for (dev = list; dev; dev = dev->next) {
			check_config(dev);
		}
	}
	print_modem_info(format, list);
	usb_free_list(list);
	return 0;
}

int modem_main(int argc, char **argv)
{
	char *tty_port = usb_modem_get_port("wan0_", MODEM_PORT_UI);
	if (!*tty_port) {
		printf("Can't find port\n");
		return 1;
	}

	int fd = port_open(tty_port);
	if (fd < 0) {
		printf("Can't open port %s\n", tty_port);
		return 1;
	}

	// workaround for disabled echo, need to redo answer parsing
	atcommand(fd, "ATE1", 0, 0, 0);
	if (atcommand(fd, "AT", 0, 0, 0)) {
		close(fd);
		printf("%s doesn't response to AT commands\n", tty_port);
		return 1;
	}

	modem_signal_t *sig = modem_get_signal(fd);
	modem_netinfo_t *netinfo = modem_get_netinfo(fd);
	char *modem_s = modem_get_model(fd);
	modem_traf_t *traf = modem_get_traffic_info(fd);
	modem_sysinfo_t *sysinfo = modem_get_sysinfo(fd);

	printf("Modem: %s\n", modem_s);
	if (netinfo) {
		printf("Operator: %s\n", (netinfo->stat == 1 || netinfo->stat == 5) ? netinfo->operrat : netinfo->stat_s);
		if (*netinfo->mcc && *netinfo->mnc) {
			printf("MCC: %s, MNC: %s", netinfo->mcc, netinfo->mnc);
			if (netinfo->lac && netinfo->ci)
				printf(", LAC: %X, CellID: %X", netinfo->lac, netinfo->ci);
			printf("\n");
		}
	free(netinfo);
	}
	printf("Signal: ");
	if (sig->rssi) {
		printf("RSSI %d dBm", sig->rssi);
		if (sig->rscp)
			printf(", RSCP %d dBm, Ec/Io %d dB", sig->rscp, sig->ecio);
		else if (sig->rsrp)
			printf(", RSRP %d dBm, RSRQ %d dB", sig->rsrp, sig->rsrq);
	}
	printf("\n");
	free(sig);
	if (sysinfo) {
		printf("System: %s%s%s%s\n", sysinfo->sysmode_name, *sysinfo->submode_name ? " (" : "", sysinfo->submode_name, *sysinfo->submode_name ? ")" : "");
		free(sysinfo);
	}
	if (traf) {
		int lsecs = traf->last_ds_time;
		int tsecs = traf->total_ds_time;
		printf("Last Session: %d day%s %02d:%02d:%02d %llu B %llu B\nTotal Sessions: %d day%s %02d:%02d:%02d %llu B %llu B\n",
		       lsecs / 86400, (lsecs / 86400 == 1 ? "" : "s"), (lsecs / 3600) % 24,
		       (lsecs / 60) % 60, lsecs % 60, traf->last_rx_flow, traf->last_tx_flow,
		       tsecs / 86400, (tsecs / 86400 == 1 ? "" : "s"), (tsecs / 3600) % 24,
		       (tsecs / 60) % 60, tsecs % 60, traf->total_rx_flow, traf->total_tx_flow);
		free(traf);
	}

	close(fd);

	return 0;
}
